"""Exponential distribution"""

from collections import namedtuple

import numpy as np
from scipy.stats import expon

from xgboost_distribution.compat import linalg_solve
from xgboost_distribution.distributions.base import BaseDistribution
from xgboost_distribution.distributions.utils import check_all_ge_zero, safe_exp

Params = namedtuple("Params", ("scale"))


class Exponential(BaseDistribution):
    """Exponential distribution with log score

    Definition:

        f(x) = 1 / scale * e^(-x / scale)

    We reparameterize scale -> log(scale) = a to ensure scale >= 0. Gradient:

        d/da -log[f(x)] = d/da -log[1/e^a e^(-x / e^a)]
                        = 1 - x e^-a
                        = 1 - x / scale

    The Fisher information = 1 / scale^2, when reparameterized:

        1 / scale^2 = I ( d/d(scale) log(scale) )^2 = I ( 1/ scale )^2

    Hence we find: I = 1

    """

    @property
    def params(self):
        return Params._fields

    def check_target(self, y):
        check_all_ge_zero(y)

    def gradient_and_hessian(self, y, params, natural_gradient=True):
        """Gradient and diagonal hessian"""

        (scale,) = self.predict(params)

        grad = np.zeros(shape=(len(y), 1), dtype="float32")
        grad[:, 0] = 1 - y / scale

        if natural_gradient:
            fisher_matrix = np.ones(shape=(len(y), 1, 1), dtype="float32")

            grad = linalg_solve(fisher_matrix, grad)
            hess = np.ones(shape=(len(y), 1), dtype="float32")  # constant hessian
        else:
            hess = -(grad - 1)

        return grad, hess

    def loss(self, y, params):
        (scale,) = self.predict(params)
        return "Exponential-NLL", -expon.logpdf(y, scale=scale)

    def predict(self, params):
        scale = safe_exp(params)
        return Params(scale=scale)

    def starting_params(self, y):
        return Params(scale=np.log(np.mean(y)))



"""Hyperparameter tuning example"""

from sklearn.datasets import fetch_california_housing
from sklearn.metrics import make_scorer
from sklearn.model_selection import GridSearchCV

from xgboost_distribution import XGBDistribution
from xgboost_distribution.metrics import get_ll_score_func


def main():
    data = fetch_california_housing()
    X, y = data.data, data.target

    distribution = "normal"

    param_grid = {"n_estimators": [5, 10, 20], "max_depth": [1, 2, 3]}

    xgb_cv = GridSearchCV(
        XGBDistribution(distribution=distribution),
        param_grid,
        cv=5,
        scoring={
            f"{distribution}_ll": make_scorer(get_ll_score_func(distribution)),
        },
        refit=False,
    )

    xgb_cv.fit(X, y)
    print(f"CV results: {xgb_cv.cv_results_}")


if __name__ == "__main__":
    main()



import pytest

import numpy as np
import pandas as pd

from xgboost_distribution.distributions.utils import (
    check_all_ge_zero,
    check_all_gt_zero,
    check_all_integer,
    safe_exp,
)


@pytest.mark.parametrize(
    "x",
    [np.array([0, 1]), pd.Series([-1, 0, 2])],
)
def test_check_is_integer(x):
    check_all_integer(x)


@pytest.mark.parametrize(
    "x",
    [np.array([0.1, 1.2]), pd.Series([-1.1, 0.4, 2.3])],
)
def test_check_is_integer_raises(x):
    with pytest.raises(ValueError):
        check_all_integer(x)


@pytest.mark.parametrize(
    "x",
    [np.array([0, 1]), pd.Series([0, 1.2, 2])],
)
def test_check_is_ge_zero(x):
    check_all_ge_zero(x)


@pytest.mark.parametrize(
    "x",
    [np.array([-0.1, 1.2]), pd.Series([-1.1, 0.4, 2.3])],
)
def test_check_is_ge_zero_raises(x):
    with pytest.raises(ValueError):
        check_all_ge_zero(x)


@pytest.mark.parametrize(
    "x",
    [np.array([1, 2]), pd.Series([0.1, 1.2, 2.2])],
)
def test_check_is_gt_zero(x):
    check_all_gt_zero(x)


@pytest.mark.parametrize(
    "x",
    [np.array([0, 1]), pd.Series([-1.1, 0.4, 2.3])],
)
def test_check_is_gt_zero_raises(x):
    with pytest.raises(ValueError):
        check_all_gt_zero(x)


@pytest.mark.parametrize(
    "x",
    [
        np.array([100], dtype="float32"),  # will get clipped to max
        np.array([-100], dtype="float32"),  # will get clipped to min
    ],
)
def test_safe_exp(x):
    x_exp = safe_exp(x)

    # ensure no infs and zeros
    assert np.isfinite(x_exp).all()
    assert x_exp.all()

    # check stabiilty in divison setting
    large_div = np.float32(1e6) / x_exp
    assert np.isfinite(large_div).all()

    small_div = np.float32(1e-6) / x_exp
    assert small_div.all()



"""Tests against data edge cases"""

import pytest

import numpy as np

from xgboost_distribution.model import XGBDistribution


def generate_data_with_target_outliers(n_samples=30_000, n_outliers=100):
    """Generate data with small number of outliers in target

    Bulk of target is drawn from single distribution, with a *small* set of
    outliers. These inflate the variance for these data points, potentially
    causing overflow errors, if parameters are unconstrained.
    """
    y = np.concatenate(
        (
            np.random.normal(loc=3000, scale=500, size=n_samples - n_outliers),
            np.random.uniform(low=10_000, high=100_000, size=n_outliers),
        )
    )
    x = np.random.gamma(shape=2, scale=2, size=n_samples)
    return x[..., np.newaxis], y


def test_target_outlier_robust():
    X, y = generate_data_with_target_outliers()
    xgbd = XGBDistribution(
        distribution="normal",
        n_estimators=2,
    ).fit(X, y)

    # assert that we got here without issues
    assert isinstance(xgbd, XGBDistribution)


def generate_data_with_target_consts(n_samples=30_000, n_not_constant=100):
    """Generate data with constant/duplicates in target

    Having constant in target can cause variance to diminish, potentially
    causing overflow errors.
    """
    y = np.concatenate(
        (
            np.random.normal(loc=10, scale=5, size=n_not_constant),
            np.ones(shape=(n_samples - n_not_constant,)),
        )
    )
    x = np.concatenate((np.random.uniform(low=0, high=10, size=n_samples),))
    return x[..., np.newaxis], y


@pytest.mark.slow
def test_target_with_consts():
    X, y = generate_data_with_target_consts()
    xgbd = XGBDistribution(
        distribution="normal",
        n_estimators=100,
    ).fit(X, y)

    # assert that we got here without issues
    assert isinstance(xgbd, XGBDistribution)



"""XGBDistribution model"""

import importlib
import json
import os
from typing import (
    Any,
    Callable,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
    no_type_check,
)

import numpy as np
from sklearn.utils.validation import check_is_fitted
from xgboost._typing import ArrayLike
from xgboost.callback import TrainingCallback
from xgboost.config import config_context
from xgboost.core import Booster, DMatrix, _deprecate_positional_args
from xgboost.sklearn import (
    XGBModel,
    XGBRegressor,
    _wrap_evaluation_matrices,
    xgboost_model_doc,
)
from xgboost.training import train

from xgboost_distribution.distributions import get_distribution, get_distribution_doc
from xgboost_distribution.utils import to_serializable


@xgboost_model_doc(
    "Implementation of XGBoost to estimate distributions (in scikit-learn API).",
    ["estimators", "model"],
    extra_parameters=get_distribution_doc()
    + """
    natural_gradient : bool, default=True
        Whether or not natural gradients should be used.""",
)
class XGBDistribution(XGBRegressor):
    @_deprecate_positional_args
    def __init__(
        self,
        *,
        distribution: Optional[str] = None,
        natural_gradient: bool = True,
        objective: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        self.distribution = distribution or "normal"
        self.natural_gradient = natural_gradient

        if objective is not None:
            raise ValueError(
                "Please do not set objective directly! Use the `distribution` kwarg"
            )
        else:
            super().__init__(objective=None, **kwargs)

    @_deprecate_positional_args
    def fit(
        self,
        X: ArrayLike,
        y: ArrayLike,
        *,
        sample_weight: Optional[ArrayLike] = None,
        eval_set: Optional[Sequence[Tuple[ArrayLike, ArrayLike]]] = None,
        verbose: Optional[Union[bool, int]] = True,
        xgb_model: Optional[Union[Booster, str, XGBModel]] = None,
        sample_weight_eval_set: Optional[Sequence[ArrayLike]] = None,
        feature_weights: Optional[ArrayLike] = None,
    ) -> "XGBDistribution":
        with config_context(verbosity=self.verbosity):
            evals_result: TrainingCallback.EvalsLog = {}

            self._distribution = get_distribution(self.distribution)
            self._distribution.check_target(y)

            params = self.get_xgb_params()

            # we remove unexpected (i.e. not xgb native) params before fitting
            for param in ["distribution", "natural_gradient"]:
                params.pop(param)

            params["objective"] = None
            params["disable_default_eval_metric"] = True
            params["num_class"] = len(self._distribution.params)

            # we set `base_score` to zero and instead use base_margin in dmatrices
            # -> this allows different starting values for each distribution parameter
            params["base_score"] = 0.0
            self._starting_params = self._distribution.starting_params(y)

            base_margin = self._get_base_margin(len(y))
            if eval_set is not None:
                base_margin_eval_set: Optional[List[np.ndarray]] = [
                    self._get_base_margin(len(evals[1])) for evals in eval_set
                ]
            else:
                base_margin_eval_set = None

            model, _, params = self._configure_fit(
                booster=xgb_model,
                params=params,
            )

            train_dmatrix, evals = _wrap_evaluation_matrices(
                missing=self.missing,
                X=X,
                y=y,
                group=None,
                qid=None,
                sample_weight=sample_weight,
                base_margin=base_margin,
                feature_weights=feature_weights,
                eval_set=eval_set,
                sample_weight_eval_set=sample_weight_eval_set,
                base_margin_eval_set=base_margin_eval_set,
                eval_group=None,
                eval_qid=None,
                create_dmatrix=self._create_dmatrix,
                enable_categorical=self.enable_categorical,
                feature_types=self.feature_types,
            )

            self._Booster = train(
                params,
                train_dmatrix,
                num_boost_round=self.get_num_boosting_rounds(),
                evals=evals,
                early_stopping_rounds=self.early_stopping_rounds,
                evals_result=evals_result,
                obj=self._objective_func(),
                custom_metric=self._evaluation_func(),
                verbose_eval=verbose,
                xgb_model=model,
                callbacks=self.callbacks,
            )

            self._set_evaluation_result(evals_result)
            self.objective = f"distribution:{self.distribution}"

            # we set additional params needed for XGBDistribution on the Booster object,
            # in order to make use of the Booster's serialisation methods
            self._Booster.set_attr(distribution=self.distribution)
            self._Booster.set_attr(
                starting_params=json.dumps(
                    self._starting_params, default=to_serializable
                )
            )
            return self

    assert XGBModel.fit.__doc__ is not None
    fit.__doc__ = XGBModel.fit.__doc__.replace(
        "Fit gradient boosting model", "Fit gradient boosting distribution model", 1
    )

    @no_type_check
    def predict(
        self,
        X: ArrayLike,
        validate_features: bool = True,
        iteration_range: Optional[Tuple[int, int]] = None,
    ) -> Tuple[np.ndarray]:
        """Predict all params of distribution of each `X` example.

        Parameters
        ----------
        X : ArrayLike
            Feature matrix.
        validate_features : bool
            When this is True, validate that the Booster's and data's feature_names are
            identical.  Otherwise, it is assumed that the feature_names are the same.
        iteration_range :
            Specifies which layer of trees are used in prediction.  For example, if a
            random forest is trained with 100 rounds.  Specifying `iteration_range=(10,
            20)`, then only the forests built during [10, 20) (half open set) rounds are
            used in this prediction.

        Returns
        -------
        predictions : namedtuple
            A namedtuple of the distribution parameters. Each parameter is a
            numpy array of shape (n_samples,).
        """
        with config_context(verbosity=self.verbosity):
            check_is_fitted(self, attributes=("_distribution", "_starting_params"))

            base_margin = self._get_base_margin(X.shape[0])

            params = super().predict(
                X=X,
                output_margin=True,
                validate_features=validate_features,
                base_margin=base_margin,
                iteration_range=iteration_range,
            )
            return self._distribution.predict(params)

    def save_model(self, fname: Union[str, os.PathLike]) -> None:
        # self._distribution class cannot be saved by `super().save_model`, as it
        # attempts to call `json.dumps({"_distribution": self._distribution})`
        # Hence we delete, and then reinstantiate
        # (this is safe as distributions are by definition stateless)
        del self._distribution
        super().save_model(fname)
        self._distribution = get_distribution(self.distribution)

    def load_model(self, fname: Union[str, bytearray, os.PathLike]) -> None:
        super().load_model(fname)

        # See above: Currently need to reinstantiate distribution post loading
        self.distribution = self._Booster.attr("distribution")
        self._distribution = get_distribution(self.distribution)

        distribution_module = importlib.import_module(self._distribution.__module__)
        self._starting_params = distribution_module.Params(
            *json.loads(self._Booster.attr("starting_params"))
        )
        del distribution_module

    def _objective_func(
        self,
    ) -> Callable[[np.ndarray, DMatrix], Tuple[np.ndarray, np.ndarray]]:
        def obj(params: np.ndarray, data: DMatrix) -> Tuple[np.ndarray, np.ndarray]:
            y = data.get_label()
            grad, hess = self._distribution.gradient_and_hessian(
                y=y, params=params, natural_gradient=self.natural_gradient
            )

            weights = data.get_weight()
            if weights.size != 0:
                weights = weights.reshape(-1, 1)
                grad *= weights
                hess *= weights
            return grad, hess

        return obj

    def _evaluation_func(self) -> Callable[[np.ndarray, DMatrix], Tuple[str, float]]:
        def feval(params: np.ndarray, data: DMatrix) -> Tuple[str, float]:
            y = data.get_label()
            weights = data.get_weight()
            if weights.size == 0:
                weights = None

            loss_name, loss = self._distribution.loss(y=y, params=params)
            return loss_name, np.average(loss, weights=weights)

        return feval

    def _get_base_margin(self, n_samples: int) -> np.ndarray:
        return np.ones(shape=(n_samples, 1)) * np.array(self._starting_params)



import numpy as np

if np.__version__ >= "2.0.0":
    """Required as the `b` input array shape is treated differently.

    https://numpy.org/doc/stable/reference/generated/numpy.linalg.solve.html
    """

    def linalg_solve(a, b):
        x = np.linalg.solve(a, b[..., np.newaxis])
        return np.squeeze(x, axis=2)
else:
    linalg_solve = np.linalg.solve



"""Test suite for metrics"""

from xgboost_distribution.distributions import AVAILABLE_DISTRIBUTIONS
from xgboost_distribution.metrics import get_ll_score_func


def test_get_ll_score_func_distribution_exist():
    for distribution in AVAILABLE_DISTRIBUTIONS:
        score_func = get_ll_score_func(distribution=distribution)
        assert callable(score_func)



import pytest

import numpy as np
import pandas as pd

from xgboost_distribution.distributions import Exponential


@pytest.fixture
def exponential():
    return Exponential()


def test_target_validation(exponential):
    valid_target = np.array([0, 0.1, 1.4, 5.5, 12.3])
    exponential.check_target(valid_target)


@pytest.mark.parametrize(
    "invalid_target",
    [np.array([-0.1, 1.2]), pd.Series([-1.1, 0.4, 2.3])],
)
def test_target_validation_raises(exponential, invalid_target):
    with pytest.raises(ValueError):
        exponential.check_target(invalid_target)


@pytest.mark.parametrize(
    "y, params, natural_gradient, expected_grad",
    [
        (
            np.array([1, 1]),
            np.array([np.log(1), np.log(2)]),
            True,
            np.array([[0], [0.5]]),
        ),
        (
            np.array([1, 1]),
            np.array([np.log(1), np.log(2)]),
            False,
            np.array([[0], [0.5]]),
        ),
    ],
)
def test_gradient_calculation(exponential, y, params, natural_gradient, expected_grad):
    grad, _ = exponential.gradient_and_hessian(
        y, params, natural_gradient=natural_gradient
    )
    np.testing.assert_array_equal(grad, expected_grad)


def test_loss(exponential):
    loss_name, loss_values = exponential.loss(
        y=np.array([1]),
        params=np.array([np.log(1)]),
    )
    assert loss_name == "Exponential-NLL"
    np.testing.assert_array_equal(loss_values, np.array([1.0]))


@pytest.mark.parametrize(
    "y, params",
    [
        (
            np.array([34.7, 20.1], dtype="float32"),
            np.array([150, 20], dtype="float32"),
        ),
        (
            np.array([34.7, 20.1], dtype="float32"),
            np.array([-150, 20], dtype="float32"),
        ),
    ],
)
def test_overflow_stability(exponential, y, params):
    """Test stability against large/small values produced by xgboost"""

    # Scale should not have inf or zeros
    scale = exponential.predict(params)
    assert np.isfinite(scale).all()
    assert np.all(scale)
